@@ -1,5 +1,5 @@
 # -*- mode:python; coding:utf-8 -*-
-# Copyright (c) 2020 IBM Corp. All rights reserved.
+# Copyright (c) 2021, 2022 IBM Corp. All rights reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -25,6 +25,7 @@ from pathlib import Path, PurePath
 from threading import Lock
 from urllib.parse import urlparse
 
+from compliance.agent import ComplianceAgent
 from compliance.config import get_config
 from compliance.evidence import CONTENT_FLAGS, TmpEvidence, get_evidence_class
 from compliance.utils.data_parse import (
@@ -34,7 +35,8 @@ from compliance.utils.exceptions import (
     EvidenceNotFoundError,
     HistoricalEvidenceNotFoundError,
     LockerPushError,
-    StaleEvidenceError
+    StaleEvidenceError,
+    UnverifiedEvidenceError
 )
 
 import git
@@ -66,11 +68,13 @@ class Locker(object):
         self,
         name=None,
         repo_url=None,
-        branch='master',
+        branch=None,
         creds=None,
         do_push=False,
         ttl_tolerance=0,
-        gitconfig=None
+        gitconfig=None,
+        local_path=None,
+        use_extra_lockers=True
     ):
         """
         Construct and initialize the locker object.
@@ -88,6 +92,8 @@ class Locker(object):
         :param ttl_tolerance: the applied to evidence time to live.  If within
           that tolerance the evidence will be treated as stale.
         :param gitconfig: the git configuration to be applied to the locker.
+        :param local_path: a path to a local git repository.
+        :param use_extra_lockers: get evidence from configured 'extra' lockers.
         """
         self.repo_url = repo_url
         self.repo_url_with_creds = self.repo_url
@@ -107,14 +113,22 @@ class Locker(object):
                 self.repo_url_with_creds = re.sub(
                     '://', f'://{token}@', self.repo_url_with_creds
                 )
-        self.branch = branch
+        self.default_branch = self.branch = get_config().get(
+            'locker.default_branch', default='master'
+        )
+        if branch:
+            self.branch = branch
+        self._new_branch = False
         if name is not None:
             self.name = name
         elif repo_url is not None and name is None:
             self.name = repo_url.rsplit('/', 1)[1]
         elif repo_url is None and name is None:
             self.name = 'example'
-        self.local_path = str(PurePath(tempfile.gettempdir(), self.name))
+        if local_path:
+            self.local_path = local_path
+        else:
+            self.local_path = str(PurePath(tempfile.gettempdir(), self.name))
         self._do_push = do_push
         self.ttl_tolerance = ttl_tolerance
         self.gitconfig = gitconfig or {}
@@ -129,6 +143,41 @@ class Locker(object):
         self.commit_date = dt.utcnow().isoformat()
         self.forced_evidence = []
         self.dependency_rerun = []
+        if use_extra_lockers:
+            self._extra_lockers = self._get_extra_lockers()
+        else:
+            self._extra_lockers = []
+        self._clone_depth = get_config().get('locker.depth')
+        days = get_config().get('locker.shallow_days')
+        self._clone_shallow_since_days = days
+
+    @property
+    def clone_depth(self):
+        """
+        Provide the commit depth used when cloning the repository.
+
+        :returns: the clone depth.
+        """
+        return self._clone_depth
+
+    @property
+    def clone_shallow_since_days(self):
+        """
+        Provide the maximum commit age to fetch when cloning the repository.
+
+        :returns: the shallow since days.
+        """
+        return self._clone_shallow_since_days
+
+    @property
+    def head_commit_hexsha(self):
+        """
+        Get SHA of repository head commit.
+
+        :returns: 40 byte hexadecimal version of 20 byte binary SHA.
+        """
+        with self.lock:
+            return self.repo.head.commit.hexsha
 
     @property
     def touched_files(self):
@@ -157,6 +206,36 @@ class Locker(object):
                     touched_files.append(str(PurePath(f).with_name(t)))
         return touched_files
 
+    @clone_depth.setter
+    def clone_depth(self, depth):
+        self._clone_depth = depth
+
+    @clone_shallow_since_days.setter
+    def clone_shallow_since_days(self, days):
+        self._clone_shallow_since_days = days
+
+    def _get_extra_lockers(self):
+        """Get extra locker configurations."""
+        extra_lockers_cfg = get_config().get('locker.extra', [])
+        prev_repo_url = get_config().get('locker.prev_repo_url')
+        if prev_repo_url:
+            extra_lockers_cfg.append({'repo_url': prev_repo_url})
+        extra_lockers = []
+        for extra_locker_cfg in extra_lockers_cfg:
+            extra_locker = Locker(
+                branch=extra_locker_cfg.get('branch'),
+                creds=self.creds,
+                local_path=extra_locker_cfg.get('local_path'),
+                repo_url=extra_locker_cfg['repo_url'],
+                use_extra_lockers=False
+            )
+            extra_locker.clone_depth = extra_locker_cfg.get('depth')
+            days = extra_locker_cfg.get('shallow_days')
+            extra_locker.clone_shallow_since_days = days
+            extra_locker.get_locker_repo(locker=extra_locker.name)
+            extra_lockers.append(extra_locker)
+        return extra_lockers
+
     def reset_depenency_rerun(self):
         """Reset the dependency_rerun attribute to an empty list."""
         self.dependency_rerun = []
@@ -250,6 +329,14 @@ class Locker(object):
                 'ttl': evidence.ttl,
                 'description': evidence.description
             }
+            if evidence.signature:
+                metadata[evidence.name].update(
+                    {
+                        'agent_name': evidence.agent.name,
+                        'digest': evidence.digest,
+                        'signature': evidence.signature,
+                    }
+                )
             if evidence.is_empty:
                 metadata[evidence.name]['empty'] = True
             tombstones = None
@@ -452,36 +539,24 @@ class Locker(object):
 
         :returns: the populated evidence object.
         """
-        evidence = None
+        missing_errs = (EvidenceNotFoundError, HistoricalEvidenceNotFoundError)
         try:
             evidence = self._get_evidence(
                 evidence_path, ignore_ttl, evidence_dt
             )
-        except (EvidenceNotFoundError, HistoricalEvidenceNotFoundError):
-            prev_repo_url = get_config().get('locker.prev_repo_url')
-            if not (prev_repo_url and
-                    evidence_dt) or dt.utcnow().date() <= evidence_dt.date():
-                raise
-            # Try the previous locker
-            if not hasattr(self, '_prev_locker'):
-                self._prev_locker = Locker(
-                    f'{self.name}_prev',
-                    prev_repo_url,
-                    self.branch,
-                    self.creds
-                )
-            self.logger.info(
-                f'{evidence_dt.strftime("%Y-%m-%d")} historical evidence '
-                f'{evidence_path} not found in locker.  '
-                'Looking in previous locker...'
-            )
-            self._prev_locker.get_locker_repo('previous locker')
-            evidence = self._prev_locker._get_evidence(
-                evidence_path, ignore_ttl, evidence_dt
-            )
-            # Preserve previous locker for report metadata and TOC.
-            evidence.locker = self._prev_locker
-        return evidence
+            evidence.locker = self
+            return evidence
+        except missing_errs as missing_err:
+            for extra_locker in self._extra_lockers:
+                try:
+                    evidence = extra_locker._get_evidence(
+                        evidence_path, ignore_ttl, evidence_dt
+                    )
+                    evidence.locker = extra_locker
+                    return evidence
+                except missing_errs:
+                    continue
+            raise missing_err
 
     def load_content(self, evidence, ignore_ttl=False, evidence_dt=None):
         """
@@ -514,15 +589,24 @@ class Locker(object):
                     root.extend(parse_dot_key(data, evidence.part_root))
                 else:
                     root.extend(data)
-            evidence.set_content(format_json(content))
+            evidence.set_content(format_json(content), sign=False)
         else:
             evidence.set_content(
                 self._get_file_content(
                     evidence.path,
                     evidence_dt,
                     getattr(evidence, 'binary_content', False)
-                )
+                ),
+                sign=False
             )
+        ign_sig = get_config().get('locker.ignore_signatures', default=False)
+        if not ign_sig and evidence.is_signed(self):
+            if not evidence.verify_signature(self):
+                raise UnverifiedEvidenceError(
+                    f'Evidence {evidence.path} is signed but the signature '
+                    'could not be verified.',
+                    evidence
+                )
         return evidence
 
     def validate(self, evidence, ignore_ttl=False):
@@ -564,22 +648,28 @@ class Locker(object):
             self.logger.info(
                 f'Cloning {locker} {self.repo_url} to {self.local_path}...'
             )
-            kwargs = {'branch': self.branch}
-            shallow_days = get_config().get('locker.shallow_days', -1)
+            kwargs = {'branch': self.default_branch}
             addl_msg = None
-            if shallow_days >= 0:
-                since_dt = dt.utcnow() - timedelta(days=shallow_days + 1)
+            if self.clone_depth:
+                kwargs['depth'] = self.clone_depth
+            if self.clone_shallow_since_days:
+                days = self.clone_shallow_since_days + 1
+                since_dt = dt.utcnow() - timedelta(days=days)
                 since = since_dt.strftime('%Y/%m/%d')
                 kwargs['shallow_since'] = since
                 addl_msg = f'{locker.title()} contains commits since {since}'
             start = time.perf_counter()
             self.repo = git.Repo.clone_from(
-                self.repo_url_with_creds, self.local_path, **kwargs
+                self.repo_url_with_creds,
+                self.local_path,
+                single_branch=True,
+                **kwargs
             )
             duration = time.perf_counter() - start
             self.logger.info(f'{locker.title()} cloned in {duration:.3f}s')
             if addl_msg:
                 self.logger.info(addl_msg)
+        self._checkout_branch()
 
     def init_config(self):
         """Apply the git configuration."""
@@ -619,12 +709,17 @@ class Locker(object):
                 f'Syncing local locker with remote repo {self.repo_url}...'
             )
             remote.fetch()
-            remote.pull(rebase=True)
+            if not self._new_branch:
+                remote.pull(rebase=True)
             self._log_large_files()
             self.logger.info(
                 f'Pushing local locker to remote repo {self.repo_url}...'
             )
-            push_info = remote.push()[0]
+            push_info = remote.push(
+                self.branch,
+                force=get_config().get('locker.force_push', default=False),
+                set_upstream=True
+            )[0]
             if push_info.flags >= git.remote.PushInfo.ERROR:
                 raise LockerPushError(push_info)
 
@@ -750,7 +845,8 @@ class Locker(object):
         if dt:
             options['until'] = dt.strftime('%Y-%m-%d')
         try:
-            commit = next(self.repo.iter_commits(paths=path, **options))
+            with self.lock:
+                commit = next(self.repo.iter_commits(paths=path, **options))
         except StopIteration:
             pass
         return commit
@@ -813,13 +909,31 @@ class Locker(object):
         else:
             self.logger.info(f'Creating locker in {self.local_path}...')
             self.repo = git.Repo.init(self.local_path)
+        self._checkout_branch()
         self.init_config()
 
+    def _checkout_branch(self):
+        if self.repo.active_branch.name == self.branch:
+            return
+        try:
+            self.repo.git.checkout(self.branch)
+        except git.exc.GitCommandError:
+            self._new_branch = True
+            self.repo.git.checkout('-b', self.branch)
+
     def _get_evidence(self, evidence_path, ignore_ttl=False, evidence_dt=None):
+        agent = None
         evidence = None
         try:
             metadata = self.get_evidence_metadata(evidence_path, evidence_dt)
-            class_type, category, evidence_name = evidence_path.split('/')
+            if metadata and metadata.get('agent_name'):
+                agent = ComplianceAgent(
+                    name=metadata.get('agent_name'),
+                    use_agent_dir=evidence_path.startswith(
+                        ComplianceAgent.AGENTS_DIR
+                    )
+                )
+            class_type, category, evidence_name = evidence_path.split('/')[-3:]
             evidence_class_obj = get_evidence_class(class_type)
             evidence = evidence_class_obj(
                 evidence_name,
@@ -831,13 +945,15 @@ class Locker(object):
                     'root': metadata.get('partition_root')
                 },
                 binary_content=metadata.get('binary_content', False),
-                filtered_content=metadata.get('filtered_content', False)
+                filtered_content=metadata.get('filtered_content', False),
+                agent=agent,
+                evidence_dt=evidence_dt
             )
         except TypeError:
             ev_dt_str = (evidence_dt or dt.utcnow()).strftime('%Y-%m-%d')
             raise EvidenceNotFoundError(
                 f'Evidence {evidence_path} is not found in the locker '
-                f'for {ev_dt_str}.  It may not be a valid evidence path.'
+                f'for {ev_dt_str}. It may not be a valid evidence path.'
             )
         return self.load_content(evidence, ignore_ttl, evidence_dt)
 
